@@ -636,7 +636,58 @@ def get_annotations_dict(members):
     return None
   annots_var = members["__annotations__"]
   try:
-    annots = get_atomic_python_constant(annots_var, dict)
+    annots = get_atomic_value(annots_var)
   except ConversionError:
     return None
-  return annots
+  return annots if annots.isinstance_AnnotationsDict() else None
+
+
+class Local:
+  """A possibly annotated local variable."""
+
+  def __init__(self, node, op, typ, orig, vm):
+    self._ops = [op]
+    if typ:
+      self.typ = vm.program.NewVariable([typ], [], node)
+    else:
+      # Creating too many variables bloats the typegraph, hurting performance,
+      # so we use None instead of an empty variable.
+      self.typ = None
+    self.orig = orig
+    self.vm = vm
+
+  @property
+  def last_op(self):
+    # TODO(b/74434237): This property can be removed once the usage of it in
+    # dataclass_overlay is gone.
+    return self._ops[-1]
+
+  @property
+  def stack(self):
+    return self.vm.simple_stack(self.last_op)
+
+  def update(self, node, op, typ, orig):
+    """Update this variable's annotation and/or value."""
+    if op in self._ops:
+      return
+    self._ops.append(op)
+    if typ:
+      if self.typ:
+        self.typ.AddBinding(typ, [], node)
+      else:
+        self.typ = self.vm.program.NewVariable([typ], [], node)
+    if orig:
+      self.orig = orig
+
+  def get_type(self, node, name):
+    """Gets the variable's annotation."""
+    if not self.typ:
+      return None
+    values = self.typ.Data(node)
+    if len(values) > 1:
+      self.vm.errorlog.ambiguous_annotation(self.stack, values, name)
+      return self.vm.convert.unsolvable
+    elif values:
+      return values[0]
+    else:
+      return None
