@@ -28,6 +28,11 @@ _COMPATIBLE_BUILTINS = [
 ]
 
 
+def _is_callback_protocol(typ):
+  return (isinstance(typ, mixin.Class) and typ.is_protocol and
+          "__call__" in typ.protocol_methods)
+
+
 class AbstractMatcher(utils.VirtualMachineWeakrefMixin):
   """Matcher for abstract values."""
 
@@ -241,7 +246,8 @@ class AbstractMatcher(utils.VirtualMachineWeakrefMixin):
             left.param, other_type, subst, node, view)
         if new_subst is not None:
           subst = new_subst.copy()
-          # TODO(kramm): Can we put in something more precise?
+          # NOTE: This is pretty imprecise, there might be something better to
+          # do here.
           subst[other_type.full_name] = node.program.NewVariable([], [], node)
           return subst
         else:
@@ -313,10 +319,14 @@ class AbstractMatcher(utils.VirtualMachineWeakrefMixin):
       return subst
     elif (isinstance(other_type, typing_overlay.NoReturn) or
           isinstance(left, typing_overlay.NoReturn)):
-      # `NoReturn` can only matches itself or `abstract.TypeParameter`.
+      # `NoReturn` can only matches itself, `Any`, or `abstract.TypeParameter`.
       # For the latter case, it will be used in byte code `STORE_ANNOTATION`
       # to store the `NoReturn` annotation in a dict.
-      return subst if left == other_type else None
+      if (left == other_type or isinstance(other_type, abstract.Unsolvable) or
+          isinstance(left, abstract.Unsolvable)):
+        return subst
+      else:
+        return None
     elif isinstance(other_type, mixin.Class):
       # Accumulate substitutions in "subst", or break in case of error:
       return self._match_type_against_type(left, other_type, subst, node, view)
@@ -332,7 +342,6 @@ class AbstractMatcher(utils.VirtualMachineWeakrefMixin):
           isinstance(left, (abstract.Unknown, abstract.Unsolvable))):
       # We can match anything against unknown types, and unknown types against
       # anything.
-      # TODO(kramm): Do we want to record what we matched them against?
       assert not isinstance(other_type, abstract.ParameterizedClass)
       return subst
     elif isinstance(other_type, abstract.Empty):
@@ -376,6 +385,9 @@ class AbstractMatcher(utils.VirtualMachineWeakrefMixin):
       elif other_type.full_name in [
           "__builtin__.type", "__builtin__.object", "typing.Callable"]:
         return subst
+      elif _is_callback_protocol(other_type):
+        return self._match_type_against_callback_protocol(
+            left, other_type, subst, node, view)
       elif left.cls:
         return self._match_instance_against_type(
             left, other_type, subst, node, view)
@@ -403,10 +415,14 @@ class AbstractMatcher(utils.VirtualMachineWeakrefMixin):
           if new_subst is not None:
             return new_subst
         return None
+      elif _is_callback_protocol(other_type):
+        return self._match_type_against_callback_protocol(
+            left, other_type, subst, node, view)
+      else:
+        return None
     elif isinstance(left, dataclass_overlay.FieldInstance) and left.default:
-      default = self.vm.merge_values(left.default.data)
-      return self._match_type_against_type(
-          default, other_type, subst, node, view)
+      return self._match_all_bindings(
+          left.default, other_type, subst, node, view)
     elif isinstance(left, abstract.SimpleAbstractValue):
       return self._match_instance_against_type(
           left, other_type, subst, node, view)
@@ -438,6 +454,32 @@ class AbstractMatcher(utils.VirtualMachineWeakrefMixin):
       raise NotImplementedError("Matching not implemented for %s against %s" %
                                 (type(left), type(other_type)))
 
+  def _match_type_against_callback_protocol(
+      self, left, other_type, subst, node, view):
+    """See https://www.python.org/dev/peps/pep-0544/#callback-protocols."""
+    _, method_var = self.vm.attribute_handler.get_attribute(
+        node, other_type, "__call__")
+    if not method_var or not method_var.data or any(
+        not isinstance(v, abstract.Function) for v in method_var.data):
+      return None
+    new_substs = []
+    for expected_method in method_var.data:
+      signatures = abstract_utils.get_signatures(expected_method)
+      for sig in signatures:
+        sig = sig.drop_first_parameter()  # drop `self`
+        expected_callable = (
+            self.vm.convert.pytd_convert.signature_to_callable(sig))
+        new_subst = self._match_type_against_type(
+            left, expected_callable, subst, node, view)
+        if new_subst is not None:
+          # For a set of overloaded signatures, only one needs to match.
+          new_substs.append(new_subst)
+          break
+      else:
+        # Every method_var binding must have a matching signature.
+        return None
+    return self._merge_substs(subst, new_substs)
+
   def _mutate_type_parameters(self, params, value, subst, node):
     new_subst = {p.full_name: value.to_variable(node) for p in params}
     return self._merge_substs(subst, [new_subst])
@@ -524,13 +566,17 @@ class AbstractMatcher(utils.VirtualMachineWeakrefMixin):
                              container=None):
     """Instantiate and match an abstract value."""
     instance = left.instantiate(node, container=container)
+    return self._match_all_bindings(instance, other_type, subst, node, view)
+
+  def _match_all_bindings(self, var, other_type, subst, node, view):
+    """Matches all of var's bindings against other_type."""
     new_substs = []
-    for new_view in abstract_utils.get_views([instance], node):
+    for new_view in abstract_utils.get_views([var], node):
       # When new_view and view have entries in common, we want to use the
       # entries from the old view.
       new_view.update(view)
       new_subst = self.match_var_against_type(
-          instance, other_type, subst, node, new_view)
+          var, other_type, subst, node, new_view)
       if new_subst is not None:
         new_substs.append(new_subst)
     if new_substs:
@@ -871,7 +917,7 @@ class AbstractMatcher(utils.VirtualMachineWeakrefMixin):
       converter = self.vm.convert.pytd_convert
       for signature in abstract_method.signatures:
         callable_signature = converter.signature_to_callable(
-            signature.signature, self.vm)
+            signature.signature)
         if isinstance(callable_signature, abstract.CallableClass):
           # Prevent the matcher from trying to enforce contravariance on 'self'.
           callable_signature.formal_type_parameters[0] = (
